@@ -1,9 +1,20 @@
 /* wipe
- * $Id: wipe.c,v 1.2 2004/06/12 17:49:47 berke Exp $
- * by Berke Durak
- * Author may be contacted at 'echo berke1ouvaton2org|tr 12 @.'
  *
- * URL for wipe: http://abaababa.ouvaton.org/wipe/
+ * by Berke Durak
+ *
+ * With many thanks to the following contributors:
+ *     Jason Axley
+ *     Alexey Marinichev
+ *     Chris L. Mason
+ *     Karako Miklos
+ *     Jim Paris
+ *     Thomas Schoepf
+ *
+ * Author may be contacted at 'echo berke1lambda-diode|com 12 @.'
+ *
+ * URL for wipe:   http://lambda-diode.com/software/wipe
+ * Git repository: http://github.com/berke/wipe
+ *                 git clone https://github.com/berke/wipe.git
  *
  * Securely erase files from magnetic media
  * Based on data from article "Secure Deletion of Data from Magnetic
@@ -13,8 +24,8 @@
 
 /*** defines */
 
-#define WIPE_VERSION "0.21"
-#define WIPE_DATE "2006-09-20"
+#define WIPE_VERSION "0.22"
+#define WIPE_DATE "2010-11-07"
 #define WIPE_CVS "$Id: wipe.c,v 1.2 2004/06/12 17:49:47 berke Exp $"
 
 /* exit codes */
@@ -77,9 +88,9 @@
 #include <sys/types.h>
 #include <sys/ioctl.h>
 
-
 #include "random.h"
 #include "misc.h"
+#include "version.h"
 
 /* includes ***/
 
@@ -141,6 +152,7 @@ int o_recurse = 0;
 int o_dereference_symlinks = 0;
 int o_quick = 0;
 int o_quick_passes = QUICKPASSES;
+int o_quick_passes_set = 0;
 int o_verbose = 0;
 int o_silent = 0;
 char *o_devrandom = DEVRANDOM;
@@ -158,6 +170,7 @@ int o_lg2_buffer_size = BUFLG2;
 int o_buffer_size = 1<<BUFLG2;
 int o_wipe_length_set = 0;
 int o_wipe_exact_size = 0;
+int o_skip_passes = 0;
 
 /* End of Options ***/
 
@@ -227,7 +240,7 @@ inline static void fill_random_from_table (char *b,
 
 inline static void fill_random (char *b, int n)
 {
-    rand_Fill (b, n);
+    rand_Fill ((u8 *) b, n);
 }
 
 /* fill_random ***/
@@ -560,6 +573,105 @@ static int wipe_filename_and_remove (char *fn)
 
 /* wipe_filename_and_remove ***/
 
+#include <sys/time.h>
+
+static double
+get_time_of_day(void)
+{
+    struct timeval tv;
+    gettimeofday(&tv, 0);
+    return (tv.tv_usec * 1e-6 + tv.tv_sec);
+}
+
+static double eta_start_time;
+
+static void
+eta_begin()
+{
+    eta_start_time = get_time_of_day();
+}
+
+static void
+eta_progress(char *buf, size_t bufsiz, double frac)
+{
+    double now;
+    double elapsed;
+    double total;
+    double seconds_per_year;
+    double remaining;
+    long lsec;
+    int sec;
+    long lmin;
+    int min;
+    long lhours;
+    int hours;
+    long ldays;
+    int days;
+    int weeks;
+    
+    now = get_time_of_day();
+    elapsed = now - eta_start_time;
+
+    buf[0] = '\0';
+    if (elapsed < 5 && frac < 0.1)
+        return;
+
+    total = elapsed / frac;
+    seconds_per_year = 365.25 * 24 * 60 * 60;
+    remaining = total - elapsed;
+    if (remaining > seconds_per_year)
+    {
+        snprintf(buf, bufsiz, "  ETA %g years", remaining / seconds_per_year);
+        return;
+    }
+    lsec = remaining;
+    sec = lsec % 60;
+    lmin = lsec / 60;
+    min = lmin % 60;
+    lhours = lmin / 60;
+    hours = lhours % 24;
+    ldays = lhours / 24;
+    days = ldays % 7;
+    weeks = ldays / 7;
+    if (weeks > 0)
+    {
+        snprintf(buf, bufsiz, "  ETA %dw %dd", weeks, days);
+        return;
+    }
+    if (days > 0)
+    {
+        snprintf(buf, bufsiz, "  ETA %dd %dh", days, hours);
+        return;
+    }
+    if (hours > 0)
+    {
+        snprintf(buf, bufsiz, "  ETA %dh%02dm", hours, min);
+        return;
+    }
+    snprintf(buf, bufsiz, "  ETA %dm%02ds", min, sec);
+}
+
+
+static void
+pad(char *buf, size_t bufsiz)
+{
+    size_t len;
+    
+    len = strlen(buf);
+    while (len + 1 < bufsiz)
+        buf[len++] = ' ';
+    buf[len] = '\0';
+}
+
+
+static void
+backspace(char *dst, const char *src)
+{
+    while (*src++)
+        *dst++ = '\b';
+    *dst = '\0';
+}
+
 /*** dothejob -- nonrecursive wiping of a single file or device */
 
 /* determine parameters of region to be wiped
@@ -807,8 +919,13 @@ static int dothejob (char *fn)
         debugf ("buffers_to_wipe = %d, o_buffer_size = %d, wi.n_passes = %d",
                 buffers_to_wipe, o_buffer_size, wi.n_passes);
 
+	if (o_skip_passes > 0) {
+	    printf ("\rSkip first %d pass(es)\n", o_skip_passes);
+	}
+
         /* do the passes */
-        for (i = 0; i<wi.n_passes; i++) {
+        eta_begin();
+        for (i = o_skip_passes; i<wi.n_passes; i++) {
             ssize_t wr;
 
             if (!o_silent) {
@@ -831,8 +948,20 @@ static int dothejob (char *fn)
                 if (!o_silent) {
                     t = time (0);
                     if ((bpi && (t-lt)) || ((t-lt>2) && j<(buffers_to_wipe>>1))) {
-                        fprintf (stderr,
-                                "[%8ld / %8ld]\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b", (long) j, (long) buffers_to_wipe);
+                        char buf1[30];
+                        char buf1_bs[sizeof (buf1)];
+                        char buf2[18];
+                        char buf2_bs[sizeof(buf2)];
+                        snprintf(buf1, sizeof(buf1),
+                                "[%8ld / %8ld]", (long) j, (long)buffers_to_wipe);
+                        backspace(buf1_bs, buf1);
+                        eta_progress(buf2, sizeof(buf2),
+                            ((double)i + ((double)j / buffers_to_wipe)) / wi.n_passes);
+                        if (buf2[0])
+                            pad(buf2, sizeof(buf2));
+                        backspace(buf2_bs, buf2);
+                        fprintf(stderr, "%s%s%s%s", buf1, buf2, buf2_bs,
+                            buf1_bs);
                         fflush (stderr);
                         lt = t;
                         bpi = 1;
@@ -1087,13 +1216,12 @@ void banner ()
 {
     fprintf (stderr, "This is wipe version " WIPE_VERSION ".\n"
             "\n"
-            "Author: Berke Durak.\n"
-            "Author's e-mail address: echo berke1ouvaton2org|tr 12 @.\n"
-            "Web site: http://abaababa.ouvaton.org/wipe/\n"
-            "\n"
-            "Release date: " WIPE_DATE "\n"
-            "Compiled: " __DATE__ "\n"
-            "CVS stuff: " WIPE_CVS "\n"
+            "Author:                  Berke Durak.\n"
+            "Author's e-mail address: echo berke1lambda-diode2com|tr 12 @.\n"
+            "Web site:                http://lambda-diode.com/software/wipe/\n"
+            "Release date:            " WIPE_DATE "\n"
+            "Compiled:                " __DATE__ "\n"
+            "Git version:             " WIPE_GIT "\n"
             "\n"
             "Based on data from \"Secure Deletion of Data from Magnetic and Solid-State\n"
             "Memory\" by Peter Gutmann <pgut001@cs.auckland.ac.nz>.\n");
@@ -1101,7 +1229,7 @@ void banner ()
 
 /* banner ***/
 
-#define OPTSTR "DfhvrqspciR:S:M:kFZl:o:b:Q:T:P:e"
+#define OPTSTR "X:DfhvrqspciR:S:M:kFZl:o:b:Q:T:P:e"
 
 /*** reject and usage */
 
@@ -1160,7 +1288,8 @@ void usage (void)
         "\t\t-T <tries> Set maximum number of tries for free\n"
         "\t\t\tfilename search; default is 10\n"
         "\t\t-v Show version information\n"
-        "\t\t-Z Do not attempt to wipe file size\n",
+        "\t\t-Z Do not attempt to wipe file size\n"
+        "\t\t-X <number> Skip this number of passes (useful for continuing a wiping operation)\n",
     progname
         );
 
@@ -1239,6 +1368,11 @@ int main (int argc, char **argv)
         if (c<0) break;
 
         switch (c) {
+	    case 'X': o_skip_passes = atoi(optarg);
+	        if (o_skip_passes <= 0) {
+                    reject ("number of skipped passes must be strictly positive");
+                }
+                break;
             case 'c': o_dochmod = 1; break;
             case 'D': o_dereference_symlinks = 1; break;
             case 'e': o_wipe_exact_size = 1; break;
@@ -1283,8 +1417,7 @@ int main (int argc, char **argv)
                                   }
                                   break;
             case 'Q':   o_quick_passes = atoi (optarg);
-                        if (o_quick_passes <= 0)
-                            reject ("number of quick passes must be strictly positive");
+                        o_quick_passes_set = 1;
                         break;
             case 'S':
                         if (optarg[1]) 
@@ -1327,6 +1460,10 @@ int main (int argc, char **argv)
         }
     }
 
+    if (o_quick_passes_set && !o_quick) {
+        reject ("option -Q useless without -q");
+    }
+
     if (optind >= argc) reject ("wrong number of arguments");
 
     if (o_recurse && o_dereference_symlinks) reject ("options -D and -r are mutually exclusive");
